import sys
import argparse
import time
import logging
import signal
from pkg_resources import get_distribution

from xproc import meminfo, vmstat, load, slabinfo
from xproc.util import grouper

logger = logging.getLogger("xproc.console")


def setup_logger(level=logging.INFO):
    handler = logging.StreamHandler()    # stderr
    handler.setLevel(level)
    handler.setFormatter(
        logging.Formatter(fmt="%(message)s", datefmt='%H:%M:%S'))
    console = logging.getLogger("xproc.console")
    console.addHandler(handler)
    console.setLevel(level)


def signal_handler(sig, frame):
    logger.debug("SIG: %s, frame: %s", sig, str(frame))
    logger.info("")
    sys.exit(0)


_SUB_CMD = "sub_cmd"
_CMD_MEM = ["memory", "mem"]
_CMD_VMSTAT = ["vmstat"]
_CMD_PS = ["process", "ps"]
_CMD_VER = ["version"]
_CMD_LOAD = ["load"]
# _CMD_SLABINFO = ["slabinfo"]


def parse_argv() -> argparse.Namespace:
    argv = argparse.ArgumentParser("xproc", add_help=False)

    sub_parsers = argv.add_subparsers(required=True,
                                      dest=_SUB_CMD,
                                      help="sub commands")

    sub_parsers.add_parser("version", help="Show %(prog)s version")

    ps_parser = sub_parsers.add_parser("process",
                                       aliases=["ps"],
                                       help="TODO process subcommand")
    ps_parser.add_argument("-P", "--pid", type=int, required=True, help="PID")

    mem_parser = sub_parsers.add_parser("memory",
                                        aliases=["mem"],
                                        help="memory subcommand")
    mem_parser.add_argument("--list",
                            action="store_true",
                            help="List Column Names")
    mem_parser.add_argument("-e",
                            "--extra",
                            action="append",
                            type=str,
                            help="Append Memory Column")
    mem_parser.add_argument("interval", nargs='?', default=1, type=int)
    mem_parser.add_argument("count", nargs='?', default=5, type=int)

    # out_group = mem_parser.add_mutually_exclusive_group()
    # out_group.add_argument("--text", action="store_false")
    # out_group.add_argument("--pretty", action="store_true")
    # out_group.add_argument("--csv",
    #                        type=argparse.FileType("w", encoding="utf-8"))
    vmstat_parser = sub_parsers.add_parser("vmstat", help="vmstat subcommand")
    vmstat_parser.add_argument("--list",
                               action="store_true",
                               help="List Column Names")
    vmstat_parser.add_argument("-e",
                               "--extra",
                               action="append",
                               type=str,
                               help="Append VMStat Column")
    vmstat_parser.add_argument("interval", nargs='?', default=1, type=int)
    vmstat_parser.add_argument("count", nargs='?', default=5, type=int)

    load_parser = sub_parsers.add_parser("load", help="vmstat subcommand")
    load_parser.add_argument("interval", nargs='?', default=1, type=int)
    load_parser.add_argument("count", nargs='?', default=-1, type=int)

    # slab_parser = sub_parsers.add_parser("slabinfo",
    #                                      help="slabinfo subcommand")
    # slab_parser.add_argument("--sort",
    #                          "-s",
    #                          type=str,
    #                          choices=["a", "o", "v", "l", "s"],
    #                          help="""
    #                             a: active_objs
    #                             o: num_objs
    #                             v: active_slabs
    #                             l: num_slabs
    #                             s: objsize
    #                          """)
    # slab_parser.add_argument("--top",
    #                          type=int,
    #                          default=10,
    #                          help="Top N(default=10)")
    try:
        parsed = argv.parse_args()
    except Exception:
        argv.print_help()
        sys.exit(1)
    logger.debug("parsed argv: %s", parsed)
    return parsed


def list_memory_available_column_names():
    minfo = meminfo.MemoryInfo()
    for i in grouper(5, minfo.names()):
        logger.info("\t%s", ' '.join(i))


def list_vmstat_available_column_names():
    for i in grouper(6, vmstat.SUPPORT_VMSATA_NAMES.keys()):
        logger.info("\t%s", ' '.join(i))


def show_version():
    print(get_distribution('xproc'))


def show_memory(option: argparse.Namespace):
    setup_logger()
    logger.debug("%s", option)
    if option.list:
        return list_memory_available_column_names()
    count = option.count
    interval = max(option.interval, 1)
    extras = []
    if option.extra:
        for item in option.extra:
            extras.extend([i.strip() for i in item.split(",")])
    loop = 0
    while count != 0:
        count -= 1
        try:
            minfo = meminfo.MemoryInfo()
            attrs = minfo.get_attrs(extras)
            title, data = [], []
            for attr in attrs:
                name, value = attr.name, attr.value
                val_str = str(value)
                width = max(len(name), len(val_str), 12)
                title.append(f"{name:>{width}s}")
                data.append(f"{val_str:>{width}s}")
            if loop == 0:
                logger.info(" ".join(title))
            logger.info(" ".join(data))
        finally:
            loop += 1
            time.sleep(interval)


def show_vmstat(option: argparse.Namespace):
    setup_logger()
    logger.debug("%s", option)
    if option.list:
        return list_vmstat_available_column_names()
    count = option.count
    interval = max(option.interval, 1)
    extras = []
    if option.extra:
        for item in option.extra:
            extras.extend([i.strip() for i in item.split(",")])
    else:
        extras.extend(vmstat.list_default_vmstat_names())
    loop = 0
    while count != 0:
        count -= 1
        try:
            attrs = vmstat.VMStat().get_attrs(*extras)
            title, data = [], []
            for attr in attrs:
                name, value = attr.name, attr.value
                val_str = str(value)
                width = max(len(name), len(val_str), 12)
                title.append(f"{name:>{width}s}")
                data.append(f"{val_str:>{width}s}")
            if loop == 0:
                logger.info(" ".join(title))
            logger.info(" ".join(data))
        finally:
            loop += 1
            time.sleep(interval)


def show_load(option: argparse.Namespace):
    setup_logger()
    logger.debug("%s", option)
    count = option.count
    interval = max(option.interval, 1)
    loop = 0
    while count != 0:
        count -= 1
        try:
            attrs = load.current_loadavg().get_attrs()
            title, data = [], []
            for attr in attrs:
                name, value = attr.name, attr.value
                val_str = str(value)
                width = max(len(name), len(val_str), 12)
                title.append(f"{name:>{width}s}")
                data.append(f"{val_str:>{width}s}")
            if loop == 0:
                logger.info(" ".join(title))
            logger.info(" ".join(data))
        finally:
            loop += 1
            time.sleep(interval)


# def show_slabinfo(option: argparse.Namespace):
#     setup_logger()
#     logger.debug("%s", option)
#     sort_by = option.sort
#     top_n = option.top
#     slab_info = slabinfo.current_slabinfo()


def main():
    signal.signal(signal.SIGINT, signal_handler)
    namespace = parse_argv()
    command = namespace.sub_cmd
    if command in _CMD_VER:
        show_version()
    elif command in _CMD_PS:
        pass
    elif command in _CMD_MEM:
        show_memory(namespace)
    elif command in _CMD_VMSTAT:
        show_vmstat(namespace)
    elif command in _CMD_LOAD:
        show_load(namespace)
    # elif command in _CMD_SLABINFO:
    #     show_slabinfo(namespace)



import re
from typing import List, NamedTuple

from xproc.value import (
    Attr,
    FloatValue,
    IntValue,
    current_time_attr,
)


class Loadavg(NamedTuple):
    load_1: float
    load_5: float
    load_15: float
    nr_running: int
    nr_total: int
    last_pid: int

    def get_attrs(self) -> List[Attr]:
        attrs = []
        attrs.append(current_time_attr())
        attrs.append(Attr("LOAD_1_MIN", FloatValue(self.load_1)))
        attrs.append(Attr("LOAD_5_MIN", FloatValue(self.load_5)))
        attrs.append(Attr("LOAD_15_MIN", FloatValue(self.load_15)))
        attrs.append(Attr("NR_RUNNING", IntValue(self.nr_running)))
        attrs.append(Attr("NR_TOTAL", IntValue(self.nr_total)))
        attrs.append(Attr("LAST_PID", IntValue(self.last_pid)))
        return attrs


EmptyLoadavg = Loadavg(0, 0, 0, 0, 0, 0)

LOAD_PATTERN = re.compile(r"^(?P<load_1>[0-9]+\.[0-9]+)\s+"
                          r"(?P<load_5>[0-9]+\.[0-9]+)\s+"
                          r"(?P<load_15>[0-9]+\.[0-9]+)\s+"
                          r"(?P<nr_running>[0-9]+)/(?P<nr_total>[0-9]+)\s+"
                          r"(?P<last_pid>[0-9]+)\s?$")


def current_loadavg() -> Loadavg:
    # 0.24 0.16 0.06 1/296 1968353
    with open("/proc/loadavg", mode="r", encoding="utf-8") as load:
        line = load.readlines()[0]
    match = re.match(LOAD_PATTERN, line)
    if not match:
        return EmptyLoadavg
    return Loadavg(load_1=float(match.group("load_1")),
                   load_5=float(match.group("load_5")),
                   load_15=float(match.group("load_15")),
                   nr_running=int(match.group("nr_running")),
                   nr_total=int(match.group("nr_total")),
                   last_pid=int(match.group("last_pid")))



from typing import Dict, List
import logging
from xproc.util import open_file
from xproc.value import EmptyIntAttr, Attr, IntValue, parse_int_val, current_time_attr

logger = logging.getLogger("xproc.vmstat")

NR_FREE_PAGES = "nr_free_pages"
NR_ZONE_INACTIVE_ANON = "nr_zone_inactive_anon"
NR_ZONE_ACTIVE_ANON = "nr_zone_active_anon"
NR_ZONE_INACTIVE_FILE = "nr_zone_inactive_file"
NR_ZONE_ACTIVE_FILE = "nr_zone_active_file"
NR_ZONE_UNEVICTABLE = "nr_zone_unevictable"
NR_ZONE_WRITE_PENDING = "nr_zone_write_pending"
NR_MLOCK = "nr_mlock"
NR_PAGE_TABLE_PAGES = "nr_page_table_pages"
NR_KERNEL_STACK = "nr_kernel_stack"
NR_BOUNCE = "nr_bounce"
NR_ZSPAGES = "nr_zspages"
NR_FREE_CMA = "nr_free_cma"
NUMA_HIT = "numa_hit"
NUMA_MISS = "numa_miss"
NUMA_FOREIGN = "numa_foreign"
NUMA_INTERLEAVE = "numa_interleave"
NUMA_LOCAL = "numa_local"
NUMA_OTHER = "numa_other"
NR_INACTIVE_ANON = "nr_inactive_anon"
NR_ACTIVE_ANON = "nr_active_anon"
NR_INACTIVE_FILE = "nr_inactive_file"
NR_ACTIVE_FILE = "nr_active_file"
NR_UNEVICTABLE = "nr_unevictable"
NR_SLAB_RECLAIMABLE = "nr_slab_reclaimable"
NR_SLAB_UNRECLAIMABLE = "nr_slab_unreclaimable"
NR_ISOLATED_ANON = "nr_isolated_anon"
NR_ISOLATED_FILE = "nr_isolated_file"
WORKINGSET_NODES = "workingset_nodes"
WORKINGSET_REFAULT = "workingset_refault"
WORKINGSET_ACTIVATE = "workingset_activate"
WORKINGSET_RESTORE = "workingset_restore"
WORKINGSET_NODERECLAIM = "workingset_nodereclaim"
NR_ANON_PAGES = "nr_anon_pages"
NR_MAPPED = "nr_mapped"
NR_FILE_PAGES = "nr_file_pages"
NR_DIRTY = "nr_dirty"
NR_WRITEBACK = "nr_writeback"
NR_WRITEBACK_TEMP = "nr_writeback_temp"
NR_SHMEM = "nr_shmem"
NR_SHMEM_HUGEPAGES = "nr_shmem_hugepages"
NR_SHMEM_PMDMAPPED = "nr_shmem_pmdmapped"
NR_ANON_TRANSPARENT_HUGEPAGES = "nr_anon_transparent_hugepages"
NR_UNSTABLE = "nr_unstable"
NR_VMSCAN_WRITE = "nr_vmscan_write"
NR_VMSCAN_IMMEDIATE_RECLAIM = "nr_vmscan_immediate_reclaim"
NR_DIRTIED = "nr_dirtied"
NR_WRITTEN = "nr_written"
NR_KERNEL_MISC_RECLAIMABLE = "nr_kernel_misc_reclaimable"
NR_DIRTY_THRESHOLD = "nr_dirty_threshold"
NR_DIRTY_BACKGROUND_THRESHOLD = "nr_dirty_background_threshold"
PGPGIN = "pgpgin"
PGPGOUT = "pgpgout"
PSWPIN = "pswpin"
PSWPOUT = "pswpout"
PGALLOC_DMA = "pgalloc_dma"
PGALLOC_DMA32 = "pgalloc_dma32"
PGALLOC_NORMAL = "pgalloc_normal"
PGALLOC_MOVABLE = "pgalloc_movable"
ALLOCSTALL_DMA = "allocstall_dma"
ALLOCSTALL_DMA32 = "allocstall_dma32"
ALLOCSTALL_NORMAL = "allocstall_normal"
ALLOCSTALL_MOVABLE = "allocstall_movable"
PGSKIP_DMA = "pgskip_dma"
PGSKIP_DMA32 = "pgskip_dma32"
PGSKIP_NORMAL = "pgskip_normal"
PGSKIP_MOVABLE = "pgskip_movable"
PGFREE = "pgfree"
PGACTIVATE = "pgactivate"
PGDEACTIVATE = "pgdeactivate"
PGLAZYFREE = "pglazyfree"
PGFAULT = "pgfault"
PGMAJFAULT = "pgmajfault"
PGLAZYFREED = "pglazyfreed"
PGREFILL = "pgrefill"
PGSTEAL_KSWAPD = "pgsteal_kswapd"
PGSTEAL_DIRECT = "pgsteal_direct"
PGSCAN_KSWAPD = "pgscan_kswapd"
PGSCAN_DIRECT = "pgscan_direct"
PGSCAN_DIRECT_THROTTLE = "pgscan_direct_throttle"
ZONE_RECLAIM_FAILED = "zone_reclaim_failed"
PGINODESTEAL = "pginodesteal"
SLABS_SCANNED = "slabs_scanned"
KSWAPD_INODESTEAL = "kswapd_inodesteal"
KSWAPD_LOW_WMARK_HIT_QUICKLY = "kswapd_low_wmark_hit_quickly"
KSWAPD_HIGH_WMARK_HIT_QUICKLY = "kswapd_high_wmark_hit_quickly"
PAGEOUTRUN = "pageoutrun"
PGROTATED = "pgrotated"
DROP_PAGECACHE = "drop_pagecache"
DROP_SLAB = "drop_slab"
OOM_KILL = "oom_kill"
NUMA_PTE_UPDATES = "numa_pte_updates"
NUMA_HUGE_PTE_UPDATES = "numa_huge_pte_updates"
NUMA_HINT_FAULTS = "numa_hint_faults"
NUMA_HINT_FAULTS_LOCAL = "numa_hint_faults_local"
NUMA_PAGES_MIGRATED = "numa_pages_migrated"
PGMIGRATE_SUCCESS = "pgmigrate_success"
PGMIGRATE_FAIL = "pgmigrate_fail"
COMPACT_MIGRATE_SCANNED = "compact_migrate_scanned"
COMPACT_FREE_SCANNED = "compact_free_scanned"
COMPACT_ISOLATED = "compact_isolated"
COMPACT_STALL = "compact_stall"
COMPACT_FAIL = "compact_fail"
COMPACT_SUCCESS = "compact_success"
COMPACT_DAEMON_WAKE = "compact_daemon_wake"
COMPACT_DAEMON_MIGRATE_SCANNED = "compact_daemon_migrate_scanned"
COMPACT_DAEMON_FREE_SCANNED = "compact_daemon_free_scanned"
HTLB_BUDDY_ALLOC_SUCCESS = "htlb_buddy_alloc_success"
HTLB_BUDDY_ALLOC_FAIL = "htlb_buddy_alloc_fail"
UNEVICTABLE_PGS_CULLED = "unevictable_pgs_culled"
UNEVICTABLE_PGS_SCANNED = "unevictable_pgs_scanned"
UNEVICTABLE_PGS_RESCUED = "unevictable_pgs_rescued"
UNEVICTABLE_PGS_MLOCKED = "unevictable_pgs_mlocked"
UNEVICTABLE_PGS_MUNLOCKED = "unevictable_pgs_munlocked"
UNEVICTABLE_PGS_CLEARED = "unevictable_pgs_cleared"
UNEVICTABLE_PGS_STRANDED = "unevictable_pgs_stranded"
THP_FAULT_ALLOC = "thp_fault_alloc"
THP_FAULT_FALLBACK = "thp_fault_fallback"
THP_COLLAPSE_ALLOC = "thp_collapse_alloc"
THP_COLLAPSE_ALLOC_FAILED = "thp_collapse_alloc_failed"
THP_FILE_ALLOC = "thp_file_alloc"
THP_FILE_MAPPED = "thp_file_mapped"
THP_SPLIT_PAGE = "thp_split_page"
THP_SPLIT_PAGE_FAILED = "thp_split_page_failed"
THP_DEFERRED_SPLIT_PAGE = "thp_deferred_split_page"
THP_SPLIT_PMD = "thp_split_pmd"
THP_SPLIT_PUD = "thp_split_pud"
THP_ZERO_PAGE_ALLOC = "thp_zero_page_alloc"
THP_ZERO_PAGE_ALLOC_FAILED = "thp_zero_page_alloc_failed"
THP_SWPOUT = "thp_swpout"
THP_SWPOUT_FALLBACK = "thp_swpout_fallback"
BALLOON_INFLATE = "balloon_inflate"
BALLOON_DEFLATE = "balloon_deflate"
BALLOON_MIGRATE = "balloon_migrate"
SWAP_RA = "swap_ra"
SWAP_RA_HIT = "swap_ra_hit"

SUPPORT_VMSATA_NAMES = {
    NR_FREE_PAGES: NR_FREE_PAGES,
    NR_ZONE_INACTIVE_ANON: NR_ZONE_INACTIVE_ANON,
    NR_ZONE_ACTIVE_ANON: NR_ZONE_ACTIVE_ANON,
    NR_ZONE_INACTIVE_FILE: NR_ZONE_INACTIVE_FILE,
    NR_ZONE_ACTIVE_FILE: NR_ZONE_ACTIVE_FILE,
    NR_ZONE_UNEVICTABLE: NR_ZONE_UNEVICTABLE,
    NR_ZONE_WRITE_PENDING: NR_ZONE_WRITE_PENDING,
    NR_MLOCK: NR_MLOCK,
    NR_PAGE_TABLE_PAGES: NR_PAGE_TABLE_PAGES,
    NR_KERNEL_STACK: NR_KERNEL_STACK,
    NR_BOUNCE: NR_BOUNCE,
    NR_ZSPAGES: NR_ZSPAGES,
    NR_FREE_CMA: NR_FREE_CMA,
    NUMA_HIT: NUMA_HIT,
    NUMA_MISS: NUMA_MISS,
    NUMA_FOREIGN: NUMA_FOREIGN,
    NUMA_INTERLEAVE: NUMA_INTERLEAVE,
    NUMA_LOCAL: NUMA_LOCAL,
    NUMA_OTHER: NUMA_OTHER,
    NR_INACTIVE_ANON: NR_INACTIVE_ANON,
    NR_ACTIVE_ANON: NR_ACTIVE_ANON,
    NR_INACTIVE_FILE: NR_INACTIVE_FILE,
    NR_ACTIVE_FILE: NR_ACTIVE_FILE,
    NR_UNEVICTABLE: NR_UNEVICTABLE,
    NR_SLAB_RECLAIMABLE: NR_SLAB_RECLAIMABLE,
    NR_SLAB_UNRECLAIMABLE: NR_SLAB_UNRECLAIMABLE,
    NR_ISOLATED_ANON: NR_ISOLATED_ANON,
    NR_ISOLATED_FILE: NR_ISOLATED_FILE,
    WORKINGSET_NODES: WORKINGSET_NODES,
    WORKINGSET_REFAULT: WORKINGSET_REFAULT,
    WORKINGSET_ACTIVATE: WORKINGSET_ACTIVATE,
    WORKINGSET_RESTORE: WORKINGSET_RESTORE,
    WORKINGSET_NODERECLAIM: WORKINGSET_NODERECLAIM,
    NR_ANON_PAGES: NR_ANON_PAGES,
    NR_MAPPED: NR_MAPPED,
    NR_FILE_PAGES: NR_FILE_PAGES,
    NR_DIRTY: NR_DIRTY,
    NR_WRITEBACK: NR_WRITEBACK,
    NR_WRITEBACK_TEMP: NR_WRITEBACK_TEMP,
    NR_SHMEM: NR_SHMEM,
    NR_SHMEM_HUGEPAGES: NR_SHMEM_HUGEPAGES,
    NR_SHMEM_PMDMAPPED: NR_SHMEM_PMDMAPPED,
    NR_ANON_TRANSPARENT_HUGEPAGES: NR_ANON_TRANSPARENT_HUGEPAGES,
    NR_UNSTABLE: NR_UNSTABLE,
    NR_VMSCAN_WRITE: NR_VMSCAN_WRITE,
    NR_VMSCAN_IMMEDIATE_RECLAIM: NR_VMSCAN_IMMEDIATE_RECLAIM,
    NR_DIRTIED: NR_DIRTIED,
    NR_WRITTEN: NR_WRITTEN,
    NR_KERNEL_MISC_RECLAIMABLE: NR_KERNEL_MISC_RECLAIMABLE,
    NR_DIRTY_THRESHOLD: NR_DIRTY_THRESHOLD,
    NR_DIRTY_BACKGROUND_THRESHOLD: NR_DIRTY_BACKGROUND_THRESHOLD,
    PGPGIN: PGPGIN,
    PGPGOUT: PGPGOUT,
    PSWPIN: PSWPIN,
    PSWPOUT: PSWPOUT,
    PGALLOC_DMA: PGALLOC_DMA,
    PGALLOC_DMA32: PGALLOC_DMA32,
    PGALLOC_NORMAL: PGALLOC_NORMAL,
    PGALLOC_MOVABLE: PGALLOC_MOVABLE,
    ALLOCSTALL_DMA: ALLOCSTALL_DMA,
    ALLOCSTALL_DMA32: ALLOCSTALL_DMA32,
    ALLOCSTALL_NORMAL: ALLOCSTALL_NORMAL,
    ALLOCSTALL_MOVABLE: ALLOCSTALL_MOVABLE,
    PGSKIP_DMA: PGSKIP_DMA,
    PGSKIP_DMA32: PGSKIP_DMA32,
    PGSKIP_NORMAL: PGSKIP_NORMAL,
    PGSKIP_MOVABLE: PGSKIP_MOVABLE,
    PGFREE: PGFREE,
    PGACTIVATE: PGACTIVATE,
    PGDEACTIVATE: PGDEACTIVATE,
    PGLAZYFREE: PGLAZYFREE,
    PGFAULT: PGFAULT,
    PGMAJFAULT: PGMAJFAULT,
    PGLAZYFREED: PGLAZYFREED,
    PGREFILL: PGREFILL,
    PGSTEAL_KSWAPD: PGSTEAL_KSWAPD,
    PGSTEAL_DIRECT: PGSTEAL_DIRECT,
    PGSCAN_KSWAPD: PGSCAN_KSWAPD,
    PGSCAN_DIRECT: PGSCAN_DIRECT,
    PGSCAN_DIRECT_THROTTLE: PGSCAN_DIRECT_THROTTLE,
    ZONE_RECLAIM_FAILED: ZONE_RECLAIM_FAILED,
    PGINODESTEAL: PGINODESTEAL,
    SLABS_SCANNED: SLABS_SCANNED,
    KSWAPD_INODESTEAL: KSWAPD_INODESTEAL,
    KSWAPD_LOW_WMARK_HIT_QUICKLY: KSWAPD_LOW_WMARK_HIT_QUICKLY,
    KSWAPD_HIGH_WMARK_HIT_QUICKLY: KSWAPD_HIGH_WMARK_HIT_QUICKLY,
    PAGEOUTRUN: PAGEOUTRUN,
    PGROTATED: PGROTATED,
    DROP_PAGECACHE: DROP_PAGECACHE,
    DROP_SLAB: DROP_SLAB,
    OOM_KILL: OOM_KILL,
    NUMA_PTE_UPDATES: NUMA_PTE_UPDATES,
    NUMA_HUGE_PTE_UPDATES: NUMA_HUGE_PTE_UPDATES,
    NUMA_HINT_FAULTS: NUMA_HINT_FAULTS,
    NUMA_HINT_FAULTS_LOCAL: NUMA_HINT_FAULTS_LOCAL,
    NUMA_PAGES_MIGRATED: NUMA_PAGES_MIGRATED,
    PGMIGRATE_SUCCESS: PGMIGRATE_SUCCESS,
    PGMIGRATE_FAIL: PGMIGRATE_FAIL,
    COMPACT_MIGRATE_SCANNED: COMPACT_MIGRATE_SCANNED,
    COMPACT_FREE_SCANNED: COMPACT_FREE_SCANNED,
    COMPACT_ISOLATED: COMPACT_ISOLATED,
    COMPACT_STALL: COMPACT_STALL,
    COMPACT_FAIL: COMPACT_FAIL,
    COMPACT_SUCCESS: COMPACT_SUCCESS,
    COMPACT_DAEMON_WAKE: COMPACT_DAEMON_WAKE,
    COMPACT_DAEMON_MIGRATE_SCANNED: COMPACT_DAEMON_MIGRATE_SCANNED,
    COMPACT_DAEMON_FREE_SCANNED: COMPACT_DAEMON_FREE_SCANNED,
    HTLB_BUDDY_ALLOC_SUCCESS: HTLB_BUDDY_ALLOC_SUCCESS,
    HTLB_BUDDY_ALLOC_FAIL: HTLB_BUDDY_ALLOC_FAIL,
    UNEVICTABLE_PGS_CULLED: UNEVICTABLE_PGS_CULLED,
    UNEVICTABLE_PGS_SCANNED: UNEVICTABLE_PGS_SCANNED,
    UNEVICTABLE_PGS_RESCUED: UNEVICTABLE_PGS_RESCUED,
    UNEVICTABLE_PGS_MLOCKED: UNEVICTABLE_PGS_MLOCKED,
    UNEVICTABLE_PGS_MUNLOCKED: UNEVICTABLE_PGS_MUNLOCKED,
    UNEVICTABLE_PGS_CLEARED: UNEVICTABLE_PGS_CLEARED,
    UNEVICTABLE_PGS_STRANDED: UNEVICTABLE_PGS_STRANDED,
    THP_FAULT_ALLOC: THP_FAULT_ALLOC,
    THP_FAULT_FALLBACK: THP_FAULT_FALLBACK,
    THP_COLLAPSE_ALLOC: THP_COLLAPSE_ALLOC,
    THP_COLLAPSE_ALLOC_FAILED: THP_COLLAPSE_ALLOC_FAILED,
    THP_FILE_ALLOC: THP_FILE_ALLOC,
    THP_FILE_MAPPED: THP_FILE_MAPPED,
    THP_SPLIT_PAGE: THP_SPLIT_PAGE,
    THP_SPLIT_PAGE_FAILED: THP_SPLIT_PAGE_FAILED,
    THP_DEFERRED_SPLIT_PAGE: THP_DEFERRED_SPLIT_PAGE,
    THP_SPLIT_PMD: THP_SPLIT_PMD,
    THP_SPLIT_PUD: THP_SPLIT_PUD,
    THP_ZERO_PAGE_ALLOC: THP_ZERO_PAGE_ALLOC,
    THP_ZERO_PAGE_ALLOC_FAILED: THP_ZERO_PAGE_ALLOC_FAILED,
    THP_SWPOUT: THP_SWPOUT,
    THP_SWPOUT_FALLBACK: THP_SWPOUT_FALLBACK,
    BALLOON_INFLATE: BALLOON_INFLATE,
    BALLOON_DEFLATE: BALLOON_DEFLATE,
    BALLOON_MIGRATE: BALLOON_MIGRATE,
    SWAP_RA: SWAP_RA,
    SWAP_RA_HIT: SWAP_RA_HIT,
}


def supported(name: str) -> bool:
    return name in SUPPORT_VMSATA_NAMES


class VMStat:

    def __init__(self):
        with open_file("/proc/vmstat") as vmstat:
            lines = [l.strip() for l in vmstat.readlines()]
        attrs: Dict[str, Attr] = {}
        for line in lines:
            idx = line.find(" ")
            name = line[0:idx]
            value = line[idx + 1:]
            attr = parse_int_val(name, value)
            val = attr.value
            if isinstance(val, IntValue):
                attrs[attr.name] = attr
        self._attrs = attrs

    def _get_attr(self, name: str) -> Attr:
        return self._attrs.get(name, EmptyIntAttr)

    def get_attrs(self, *names) -> List[Attr]:
        attrs = []
        attrs.append(current_time_attr())
        for name in names:
            attrs.append(self._get_attr(name))
        return attrs


def list_default_vmstat_names() -> List[str]:
    return [
        ALLOCSTALL_MOVABLE,
        ALLOCSTALL_NORMAL,
    # COMPACT_DAEMON_FREE_SCANNED,
    # COMPACT_DAEMON_MIGRATE_SCANNED,
    # COMPACT_DAEMON_WAKE,
        COMPACT_FAIL,
        COMPACT_FREE_SCANNED,
        COMPACT_ISOLATED,
        COMPACT_MIGRATE_SCANNED,
        COMPACT_STALL,
        COMPACT_SUCCESS,
    # PGMIGRATE_FAIL,
    # PGMIGRATE_SUCCESS,
    # PGSCAN_DIRECT,
    # PGSCAN_KSWAPD,
    # PGSTEAL_DIRECT,
    # PGSTEAL_KSWAPD,
    # THP_COLLAPSE_ALLOC,
    # THP_COLLAPSE_ALLOC_FAILED,
    # THP_FAULT_ALLOC,
    # THP_FAULT_FALLBACK,
    ]



import re
from typing import List, NamedTuple, Optional, OrderedDict
import collections
import operator

from xproc.util import open_file
from xproc.value import Attr, IntValue, IntUnitValue, StrValue


class AttrSlab(NamedTuple):
    name: Attr
    active_objs: Attr
    num_objs: Attr
    objsize: Attr
    objperslab: Attr
    pagesperslab: Attr
    limit: Attr
    batchcount: Attr
    sharedfactor: Attr
    active_slabs: Attr
    num_slabs: Attr
    sharedavail: Attr


class Slab(NamedTuple):
    name: str
    active_objs: int    # The number of objects that are currently active (i.e., inuse).
    num_objs: int    # The total number of allocated objects
    objsize: int    # in bytes
    objperslab: int    # The number of objects stored in each slab
    pagesperslab: int    # The number of pages allocated for each slab
    # tunables
    limit: int
    batchcount: int
    sharedfactor: int
    #slabdata
    active_slabs: int    # The number of active slabs
    num_slabs: int    # The total number of slabs
    sharedavail: int

    def to_attrslab(self) -> AttrSlab:
        return AttrSlab(
            name=Attr("NAME", StrValue(self.name)),
            active_objs=Attr("ACTIVE_OBJS", IntValue(self.active_objs)),
            num_objs=Attr("NUM_OBJS", IntValue(self.num_objs)),
            objsize=Attr("OBJ_SIZE", IntUnitValue(self.objsize, "B")),
            objperslab=Attr("OBJ_PER_SLAB", IntValue(self.objperslab)),
            pagesperslab=Attr("PAGES_PER_SLAB", IntValue(self.pagesperslab)),
            limit=Attr("LIMIT", IntValue(self.limit)),
            batchcount=Attr("BATCHCOUNT", IntValue(self.batchcount)),
            sharedfactor=Attr("SHAREDFACTOR", IntValue(self.sharedfactor)),
            active_slabs=Attr("ACTIVE_SLABS", IntValue(self.active_slabs)),
            num_slabs=Attr("NUM_SLABS", IntValue(self.num_slabs)),
            sharedavail=Attr("SHAREDAVAIL", IntValue(self.sharedavail)),
        )


# Copy from `slabtop --help`
# The following are valid sort criteria:
#  a: sort by number of active objects
#  b: sort by objects per slab
#  c: sort by cache size
#  l: sort by number of slabs
#  v: sort by number of active slabs
#  n: sort by name
#  o: sort by number of objects (the default)
#  p: sort by pages per slab
#  s: sort by object size
#  u: sort by cache utilization
_SLAB_SORT_KEYWORD = {
    "a": "active_objs",
    "o": "num_objs",
    "v": "active_slabs",
    "l": "num_slabs",
    "s": "objsize",
}


class SlabInfo:

    def __init__(self, slabs: OrderedDict[str, Slab]):
        self._slabs = slabs
        active_objs = 0
        num_objs = 0
        active_slabs = 0
        num_slabs = 0
        for slab in slabs.values():
            active_objs += slab.active_objs
            num_objs += slab.num_objs
            active_slabs += slab.active_slabs
            num_slabs += slab.num_slabs
        self._active_objs = active_objs
        self._num_objs = num_objs
        self._active_slabs = active_slabs
        self._num_slabs = num_slabs

    def active_objs(self) -> Attr:
        return Attr("ACTIVE_OBJS", IntValue(self._active_objs))

    def num_objs(self) -> Attr:
        return Attr("NUM_OBJS", IntValue(self._num_objs))

    def active_slabs(self) -> Attr:
        return Attr("ACTIVE_SLABS", IntValue(self._active_slabs))

    def num_slabs(self) -> Attr:
        return Attr("NUM_SLABS", IntValue(self._num_slabs))

    def _sort(self, bywhat: str = 'o', top: int = -1) -> List[Slab]:
        slabs = list(self._slabs.values())
        if bywhat not in _SLAB_SORT_KEYWORD:
            return slabs
        attr_name = _SLAB_SORT_KEYWORD[bywhat]
        sorted_slabs = sorted(slabs,
                              key=operator.attrgetter(attr_name),
                              reverse=True)
        if top <= 0:
            return sorted_slabs
        return sorted_slabs[0:min(top, len(sorted_slabs))]

    def sort(self, bywhat: str, top: int) -> List[AttrSlab]:
        attr_slabs = []
        for slab in self._sort(bywhat, top):
            attr_slabs.append(slab.to_attrslab())
        return attr_slabs


_SLAB_PAT = re.compile((
    r"^(?P<name>\w+)\s+(?P<active_objs>\d+)\s+(?P<num_objs>\d+)\s+"
    r"(?P<objsize>\d+)\s+(?P<objperslab>\d+)\s+(?P<pagesperslab>\d+)\s+"
    r":\s+tunables\s+(?P<limit>\d+)\s+(?P<batchcount>\d+)\s+(?P<sharedfactor>\d+)\s+"
    r":\s+slabdata\s+(?P<active_slabs>\d+)\s+(?P<num_slabs>\d+)\s+(?P<sharedavail>\d+)\s+$"
))


def _parse(line: str) -> Optional[Slab]:
    match = _SLAB_PAT.match(line)
    if not match:
        return None
    return Slab(
        match.group("name"),
        int(match.group("active_objs")),
        int(match.group("num_objs")),
        int(match.group("objsize")),
        int(match.group("objperslab")),
        int(match.group("pagesperslab")),
        int(match.group("limit")),
        int(match.group("batchcount")),
        int(match.group("sharedfactor")),
        int(match.group("active_slabs")),
        int(match.group("num_slabs")),
        int(match.group("sharedavail")),
    )


def current_slabinfo() -> SlabInfo:
    with open_file("/proc/slabinfo") as sinfo:
        lines = sinfo.readlines()
    slabs = collections.OrderedDict()
    for line in lines:
        slab = _parse(line)
        if slab:
            slabs[slab.name] = slab
    return SlabInfo(slabs)



from collections import OrderedDict
from typing import Dict, List
import pathlib
import re

from xproc.value import (
    parse_int_val,
    parse_str_val,
    parse_int_unit_val,
    DEF_FMT_1,
    DEF_FMT_2,
    Attr,
)

PID_PAT = re.compile(r"[1-9][0-9]+")

PS_NAME = "Name"
PS_UMASK = "Umask"
PS_STATE = "State"
PS_TGID = "Tgid"
PS_NGID = "Ngid"
PS_PID = "Pid"
PS_PPID = "PPid"
PS_TRACERPID = "TracerPid"
PS_UID = "Uid"
PS_GID = "Gid"
PS_FDSIZE = "FDSize"
PS_GROUPS = "Groups"
PS_NSTGID = "NStgid"
PS_NSPID = "NSpid"
PS_NSPGID = "NSpgid"
PS_NSSID = "NSsid"
PS_VMPEAK = "VmPeak"
PS_VMSIZE = "VmSize"
PS_VMLCK = "VmLck"
PS_VMPIN = "VmPin"
PS_VMHWM = "VmHWM"
PS_VMRSS = "VmRSS"
PS_RSSANON = "RssAnon"
PS_RSSFILE = "RssFile"
PS_RSSSHMEM = "RssShmem"
PS_VMDATA = "VmData"
PS_VMSTK = "VmStk"
PS_VMEXE = "VmExe"
PS_VMLIB = "VmLib"
PS_VMPTE = "VmPTE"
PS_VMSWAP = "VmSwap"
PS_HUGETLBPAGES = "HugetlbPages"
PS_COREDUMPING = "CoreDumping"
PS_THP_ENABLED = "THP_enabled"
PS_THREADS = "Threads"
PS_SIGQ = "SigQ"
PS_SIGPND = "SigPnd"
PS_SHDPND = "ShdPnd"
PS_SIGBLK = "SigBlk"
PS_SIGIGN = "SigIgn"
PS_SIGCGT = "SigCgt"
PS_CAPINH = "CapInh"
PS_CAPPRM = "CapPrm"
PS_CAPEFF = "CapEff"
PS_CAPBND = "CapBnd"
PS_CAPAMB = "CapAmb"
PS_NONEWPRIVS = "NoNewPrivs"
PS_SECCOMP = "Seccomp"
PS_SPECULATION_STORE_BYPASS = "Speculation_Store_Bypass"
PS_CPUS_ALLOWED = "Cpus_allowed"
PS_CPUS_ALLOWED_LIST = "Cpus_allowed_list"
PS_MEMS_ALLOWED = "Mems_allowed"
PS_MEMS_ALLOWED_LIST = "Mems_allowed_list"
PS_VOLUNTARY_CTXT_SWITCHES = "voluntary_ctxt_switches"
PS_NONVOLUNTARY_CTXT_SWITCHES = "nonvoluntary_ctxt_switches"

ATTR_DICT = {
    PS_NAME: [parse_str_val, DEF_FMT_1],
    PS_UMASK: [parse_str_val, DEF_FMT_1],
    PS_STATE: [parse_str_val, DEF_FMT_1],
    PS_TGID: [parse_int_val, DEF_FMT_1],
    PS_NGID: [parse_int_val, DEF_FMT_1],
    PS_PID: [parse_int_val, DEF_FMT_1],
    PS_PPID: [parse_int_val, DEF_FMT_1],
    PS_TRACERPID: [parse_int_val, DEF_FMT_1],
    PS_UID: [parse_str_val, DEF_FMT_1],
    PS_GID: [parse_str_val, DEF_FMT_1],
    PS_FDSIZE: [parse_int_val, DEF_FMT_1],
    PS_GROUPS: [parse_str_val, DEF_FMT_1],
    PS_NSTGID: [parse_int_val, DEF_FMT_1],
    PS_NSPID: [parse_int_val, DEF_FMT_1],
    PS_NSPGID: [parse_int_val, DEF_FMT_1],
    PS_NSSID: [parse_int_val, DEF_FMT_1],
    PS_VMPEAK: [parse_int_unit_val, DEF_FMT_2],
    PS_VMSIZE: [parse_int_unit_val, DEF_FMT_2],
    PS_VMLCK: [parse_int_unit_val, DEF_FMT_2],
    PS_VMPIN: [parse_int_unit_val, DEF_FMT_2],
    PS_VMHWM: [parse_int_unit_val, DEF_FMT_2],
    PS_VMRSS: [parse_int_unit_val, DEF_FMT_2],
    PS_RSSANON: [parse_int_unit_val, DEF_FMT_2],
    PS_RSSFILE: [parse_int_unit_val, DEF_FMT_2],
    PS_RSSSHMEM: [parse_int_unit_val, DEF_FMT_2],
    PS_VMDATA: [parse_int_unit_val, DEF_FMT_2],
    PS_VMSTK: [parse_int_unit_val, DEF_FMT_2],
    PS_VMEXE: [parse_int_unit_val, DEF_FMT_2],
    PS_VMLIB: [parse_int_unit_val, DEF_FMT_2],
    PS_VMPTE: [parse_int_unit_val, DEF_FMT_2],
    PS_VMSWAP: [parse_int_unit_val, DEF_FMT_2],
    PS_HUGETLBPAGES: [parse_int_unit_val, DEF_FMT_2],
    PS_COREDUMPING: [parse_int_val, DEF_FMT_1],
    PS_THP_ENABLED: [parse_int_val, DEF_FMT_1],
    PS_THREADS: [parse_int_val, DEF_FMT_1],
    PS_SIGQ: [parse_str_val, DEF_FMT_1],
    PS_SIGPND: [parse_str_val, DEF_FMT_1],
    PS_SHDPND: [parse_str_val, DEF_FMT_1],
    PS_SIGBLK: [parse_str_val, DEF_FMT_1],
    PS_SIGIGN: [parse_str_val, DEF_FMT_1],
    PS_SIGCGT: [parse_str_val, DEF_FMT_1],
    PS_CAPINH: [parse_str_val, DEF_FMT_1],
    PS_CAPPRM: [parse_str_val, DEF_FMT_1],
    PS_CAPEFF: [parse_str_val, DEF_FMT_1],
    PS_CAPBND: [parse_str_val, DEF_FMT_1],
    PS_CAPAMB: [parse_str_val, DEF_FMT_1],
    PS_NONEWPRIVS: [parse_int_val, DEF_FMT_1],
    PS_SECCOMP: [parse_int_val, DEF_FMT_1],
    PS_SPECULATION_STORE_BYPASS: [parse_str_val, DEF_FMT_1],
    PS_CPUS_ALLOWED: [parse_str_val, DEF_FMT_1],
    PS_CPUS_ALLOWED_LIST: [parse_str_val, DEF_FMT_1],
    PS_MEMS_ALLOWED: [parse_str_val, DEF_FMT_1],
    PS_MEMS_ALLOWED_LIST: [parse_str_val, DEF_FMT_1],
    PS_VOLUNTARY_CTXT_SWITCHES: [parse_int_val, DEF_FMT_1],
    PS_NONVOLUNTARY_CTXT_SWITCHES: [parse_int_val, DEF_FMT_1],
}


class PIDStatus:

    def __init__(self, pid: int):
        status_path = f"/proc/{pid}/status"
        with open(status_path, mode="r", encoding="UTF-8") as status_fp:
            lines = status_fp.readlines()
        attrs: Dict[str, Attr] = OrderedDict()
        for line in lines:
            sep_idx = line.find(":")
            name = line[0:sep_idx]
            val = line[sep_idx + 1:-1]  # exclude \n
            if name not in ATTR_DICT:
                continue
            parser, fmt = ATTR_DICT[name]
            attr = parser(name, val)
            attr.set_value_fmt(fmt)
            attrs[name] = attr
        self._attrs = attrs

    def get(self, name: str) -> str:
        if name not in self._attrs:
            return ""
        attr = self._attrs[name]
        return str(attr)

    def names(self) -> List[str]:
        return list(self._attrs.keys())


def get_all_pidstatus() -> Dict[int, PIDStatus]:
    pids = {}
    proc_path = pathlib.Path("/proc")
    for process_path in proc_path.glob("*"):
        if process_path.is_dir() and re.match(PID_PAT, process_path.name):
            pid = int(process_path.name)
            pids[pid] = PIDStatus(pid)
    return pids



from time import strftime, localtime
from abc import ABCMeta, abstractmethod
from typing import List, Union
import re

DEF_FMT_1 = "{0}"
DEF_FMT_2 = "{0}{1}"


class Value(metaclass=ABCMeta):
    """Value"""

    def __init__(self):
        self._fmt = DEF_FMT_1

    @property
    def fmt(self):
        return self._fmt

    @fmt.setter
    def fmt(self, value):
        self._fmt = value

    @abstractmethod
    def value(self) -> Union[int, str]:
        pass


class StrValue(Value):

    def __init__(self, val: str):
        super().__init__()
        self.val = val

    def value(self) -> str:
        return self.val

    def __repr__(self) -> str:
        return self.fmt.format(self.val)


class IntValue(Value):

    def __init__(self, val: int):
        super().__init__()
        self.val = val

    def value(self) -> int:
        return self.val

    def __repr__(self) -> str:
        return self.fmt.format(self.value())


class FloatValue(Value):

    def __init__(self, val: float):
        super().__init__()
        self.val = val

    def value(self) -> float:
        return self.val

    def __repr__(self) -> str:
        return self.fmt.format(self.value())


class IntUnitValue(IntValue):

    def __init__(self, val: int, unit: str):
        super().__init__(val)
        self._unit = unit
        self._fmt = DEF_FMT_2

    def unit(self) -> str:
        return self._unit

    def __repr__(self) -> str:
        return self.fmt.format(self.value(), self.unit())


class ListValue(metaclass=ABCMeta):
    """List Value"""

    @abstractmethod
    def value(self) -> List:
        pass


class ListIntValue(ListValue):

    def __init__(self, val: List[int]):
        super().__init__()
        self._val = val
        self.fmt = ""    # placeholder

    def value(self) -> List[int]:
        return self._val

    def __repr__(self) -> str:
        return " ".join([str(i) for i in self._val])


ValueType = Union[IntValue, IntUnitValue, StrValue, FloatValue, ListIntValue]


class Attr:

    def __init__(self, name: str, value: ValueType):
        self._name = name
        self._value = value

    def __repr__(self) -> str:
        return f"{self._name}:{self._value}"

    @property
    def name(self) -> str:
        return self._name

    @property
    def value(self) -> ValueType:
        return self._value

    def set_value_fmt(self, fmt: str):
        self._value.fmt = fmt


EmptyValue = StrValue("")
EmptyIntValue = IntValue(0)
EmptyIntUnitValue = IntUnitValue(0, "kB")
EmptyAttr = Attr("EmptyAttr", EmptyValue)
EmptyIntAttr = Attr("EmptyAttr", EmptyIntValue)


def parse_int_val(name: str, value: str) -> Attr:
    return Attr(name, IntValue(int(value.strip())))


def parse_int_unit_val(name: str, value: str) -> Attr:
    values = value.strip().split(" ")
    return Attr(name, IntUnitValue(int(values[0]), values[1]))


def parse_str_val(name: str, value: str) -> Attr:
    return Attr(name, StrValue(value.strip()))


def parse_list_int_val(name: str, value: str) -> Attr:
    vals = []
    for i in re.split(r"\s+", value):
        if i:
            vals.append(int(i))
    return Attr(name, ListIntValue(vals))


def current_time_attr() -> Attr:
    return Attr("TIME", StrValue(strftime("%H:%M:%S", localtime())))



import itertools


def grouper(size: int, iterable):    # Copy from stackoverflow
    aiter = iter(iterable)
    while True:
        chunk = tuple(itertools.islice(aiter, size))
        if not chunk:
            return
        yield chunk


def open_file(path: str, mode: str = "r"):
    return open(path, mode=mode, encoding="utf-8")



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['xproc']

package_data = \
{'': ['*']}

entry_points = \
{'console_scripts': ['xproc = xproc.console:main']}

setup_kwargs = {
    'name': 'xproc',
    'version': '1.3.0',
    'description': 'Linux Proc File System Snooper',
    'long_description': '# [xproc](https://github.com/hungrybirder/xproc)\n\n## Install\n\n```bash\npip install -U xproc\n```\n\n## Support commands\n\n*   `xproc mem` or `xproc memory`\n\n```bash\nxproc mem\n        TIME       KERNEL         USER      MemFree     MemTotal\n    21:19:37    1345876kB   13143712kB   10079620kB   24624680kB\n    21:19:38    1345964kB   13143796kB   10079612kB   24624680kB\n    21:19:39    1345964kB   13143804kB   10079612kB   24624680kB\n    21:19:40    1345964kB   13143808kB   10079644kB   24624680kB\n    21:19:41    1345964kB   13143808kB   10079580kB   24624680kB\n```\n\n```bash\nxproc mem -e "Active,Inactive,Active(anon),Inactive(anon),Active(file),Inactive(file)"\n        TIME       Active     Inactive Active(anon) Inactive(anon) Active(file) Inactive(file)\n    21:21:19    4397264kB    8727816kB     173164kB          128kB    4224100kB      8727688kB\n    21:21:21    4397404kB    8727816kB     173300kB          128kB    4224104kB      8727688kB\n    21:21:23    4397404kB    8727816kB     173300kB          128kB    4224104kB      8727688kB\n    21:21:25    4397404kB    8727816kB     173300kB          128kB    4224104kB      8727688kB\n    21:21:27    4397404kB    8727816kB     173300kB          128kB    4224104kB      8727688kB\n```\n\n*   `xproc vmstat`\n\n```bash\nxproc vmstat 1 5\n        TIME allocstall_movable allocstall_normal compact_fail compact_free_scanned compact_isolated compact_migrate_scanned compact_stall compact_success\n    10:15:01              40908                55            0               108344            75443                   43462            56              56\n    10:15:02              40908                55            0               108344            75443                   43462            56              56\n    10:15:03              40908                55            0               108344            75443                   43462            56              56\n    10:15:04              40908                55            0               108344            75443                   43462            56              56\n    10:15:05              40908                55            0               108344            75443                   43462            56              56\n```\n\n*   `xproc load`\n\n```bash\nxproc load 1 3\n        TIME   LOAD_1_MIN   LOAD_5_MIN  LOAD_15_MIN   NR_RUNNING     NR_TOTAL     LAST_PID\n    14:28:34         0.07         0.18         0.15            1          316      2073102\n    14:28:35         0.07         0.18         0.15            1          316      2073102\n    14:28:36         0.07         0.18         0.15            1          316      2073102\n```\n',
    'author': 'liyong',
    'author_email': 'hungrybirder@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/hungrybirder/xproc',
    'packages': packages,
    'package_data': package_data,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



from typing import NamedTuple

from xproc.util import open_file


class Uptime(NamedTuple):
    since_boot_in_seconds: float
    idle_seconds: float


def current_uptime() -> Uptime:
    with open_file("/proc/uptime") as uptime:
        line = uptime.readlines()[0].strip()
    units = line.split(" ")
    return Uptime(float(units[0]), float(units[1]))



import re
from collections import OrderedDict, defaultdict
from typing import Dict, List, NamedTuple, Tuple
import logging
from xproc.util import open_file

from xproc.value import (
    IntUnitValue,
    parse_int_unit_val,
    parse_int_val,
    DEF_FMT_1,
    DEF_FMT_2,
    Attr,
    EmptyAttr,
    current_time_attr,
)

logger = logging.getLogger("xproc.meminfo")

MEMTOTAL = "MemTotal"
MEMFREE = "MemFree"
MEMAVAILABLE = "MemAvailable"
BUFFERS = "Buffers"
CACHED = "Cached"
SWAPCACHED = "SwapCached"
ACTIVE = "Active"
INACTIVE = "Inactive"
ACTIVEANON = "Active(anon)"
INACTIVEANON = "Inactive(anon)"
ACTIVEFILE = "Active(file)"
INACTIVEFILE = "Inactive(file)"
MLOCKED = "Mlocked"
SWAPTOTAL = "SwapTotal"
SWAPFREE = "SwapFree"
DIRTY = "Dirty"
WRITEBACK = "Writeback"
ANONPAGES = "AnonPages"
MAPPED = "Mapped"
SHMEM = "Shmem"
KRECLAIMABLE = "KReclaimable"
SLAB = "Slab"
SRECLAIMABLE = "SReclaimable"
SUNRECLAIM = "SUnreclaim"
KERNELSTACK = "KernelStack"
PAGETABLES = "PageTables"
NFS_UNSTABLE = "NFS_Unstable"
BOUNCE = "Bounce"
WRITEBACKTMP = "WritebackTmp"
COMMITLIMIT = "CommitLimit"
COMMITTED_AS = "Committed_AS"
VMALLOCTOTAL = "VmallocTotal"
VMALLOCUSED = "VmallocUsed"
VMALLOCCHUNK = "VmallocChunk"
HARDWARECORRUPTED = "HardwareCorrupted"
ANONHUGEPAGES = "AnonHugePages"
SHMEMHUGEPAGES = "ShmemHugePages"
SHMEMPMDMAPPED = "ShmemPmdMapped"
HUGEPAGES_TOTAL = "HugePages_Total"
HUGEPAGES_FREE = "HugePages_Free"
HUGEPAGES_RSVD = "HugePages_Rsvd"
HUGEPAGES_SURP = "HugePages_Surp"
HUGEPAGESIZE = "Hugepagesize"
HUGETLB = "Hugetlb"
DIRECTMAP4K = "DirectMap4k"
DIRECTMAP2M = "DirectMap2M"
DIRECTMAP1G = "DirectMap1G"

ATTR_DICT = {
    MEMTOTAL: [parse_int_unit_val, DEF_FMT_2],
    MEMFREE: [parse_int_unit_val, DEF_FMT_2],
    MEMAVAILABLE: [parse_int_unit_val, DEF_FMT_2],
    BUFFERS: [parse_int_unit_val, DEF_FMT_2],
    CACHED: [parse_int_unit_val, DEF_FMT_2],
    SWAPCACHED: [parse_int_unit_val, DEF_FMT_2],
    ACTIVE: [parse_int_unit_val, DEF_FMT_2],
    INACTIVE: [parse_int_unit_val, DEF_FMT_2],
    ACTIVEANON: [parse_int_unit_val, DEF_FMT_2],
    INACTIVEANON: [parse_int_unit_val, DEF_FMT_2],
    ACTIVEFILE: [parse_int_unit_val, DEF_FMT_2],
    INACTIVEFILE: [parse_int_unit_val, DEF_FMT_2],
    MLOCKED: [parse_int_unit_val, DEF_FMT_2],
    SWAPTOTAL: [parse_int_unit_val, DEF_FMT_2],
    SWAPFREE: [parse_int_unit_val, DEF_FMT_2],
    DIRTY: [parse_int_unit_val, DEF_FMT_2],
    WRITEBACK: [parse_int_unit_val, DEF_FMT_2],
    ANONPAGES: [parse_int_unit_val, DEF_FMT_2],
    MAPPED: [parse_int_unit_val, DEF_FMT_2],
    SHMEM: [parse_int_unit_val, DEF_FMT_2],
    KRECLAIMABLE: [parse_int_unit_val, DEF_FMT_2],
    SLAB: [parse_int_unit_val, DEF_FMT_2],
    SRECLAIMABLE: [parse_int_unit_val, DEF_FMT_2],
    SUNRECLAIM: [parse_int_unit_val, DEF_FMT_2],
    KERNELSTACK: [parse_int_unit_val, DEF_FMT_2],
    PAGETABLES: [parse_int_unit_val, DEF_FMT_2],
    NFS_UNSTABLE: [parse_int_unit_val, DEF_FMT_2],
    BOUNCE: [parse_int_unit_val, DEF_FMT_2],
    WRITEBACKTMP: [parse_int_unit_val, DEF_FMT_2],
    COMMITLIMIT: [parse_int_unit_val, DEF_FMT_2],
    COMMITTED_AS: [parse_int_unit_val, DEF_FMT_2],
    VMALLOCTOTAL: [parse_int_unit_val, DEF_FMT_2],
    VMALLOCUSED: [parse_int_unit_val, DEF_FMT_2],
    VMALLOCCHUNK: [parse_int_unit_val, DEF_FMT_2],
    HARDWARECORRUPTED: [parse_int_unit_val, DEF_FMT_2],
    ANONHUGEPAGES: [parse_int_unit_val, DEF_FMT_2],
    SHMEMHUGEPAGES: [parse_int_unit_val, DEF_FMT_2],
    SHMEMPMDMAPPED: [parse_int_unit_val, DEF_FMT_2],
    HUGEPAGES_TOTAL: [parse_int_val, DEF_FMT_1],
    HUGEPAGES_FREE: [parse_int_val, DEF_FMT_1],
    HUGEPAGES_RSVD: [parse_int_val, DEF_FMT_1],
    HUGEPAGES_SURP: [parse_int_val, DEF_FMT_1],
    HUGEPAGESIZE: [parse_int_unit_val, DEF_FMT_2],
    HUGETLB: [parse_int_unit_val, DEF_FMT_2],
    DIRECTMAP4K: [parse_int_unit_val, DEF_FMT_2],
    DIRECTMAP2M: [parse_int_unit_val, DEF_FMT_2],
    DIRECTMAP1G: [parse_int_unit_val, DEF_FMT_2],
}


class MemoryInfo:

    def __init__(self, path="/proc/meminfo"):
        with open(path, mode="r", encoding="utf-8") as meminfo:
            lines = meminfo.readlines()
        attrs: Dict[str, Attr] = OrderedDict()
        for line in lines:
            sep_idx = line.find(":")
            name = line[0:sep_idx]
            val = line[sep_idx + 1:-1]    # exclude \n
            if name not in ATTR_DICT:
                continue
            parser, fmt = ATTR_DICT[name]
            attr = parser(name, val)
            attr.set_value_fmt(fmt)
            attrs[name] = attr
        self._attrs = attrs
        kernel = self._get_kernel_used_mem()
        self._attrs[kernel.name] = kernel
        user = self._get_user_used_mem()
        self._attrs[user.name] = user

    def _get_attr(self, name: str) -> Attr:
        if name not in self._attrs:
            return EmptyAttr
        return self._attrs[name]

    def get_attr_int_value(self, name: str) -> int:
        attr = self._get_attr(name)
        val = attr.value.value()
        if isinstance(val, int):
            return val
        return 0

    def names(self) -> List[str]:
        return list(self._attrs.keys())

    def _get_kernel_used_mem(self) -> Attr:
        """
        From http://linuxperf.com/?cat=7
        Kernel Used Memory includes:
            1. Slab
            2. VmallocUsed
            3. PageTables
            4. KernelStack
            5. HardwareCorrupted
            6. Bounce
            7. X, alloc_pages/__get_free_pages, but not in /proc/meminfo
        """
        slab = self._get_attr(SLAB)
        vmalloc_used = self._get_attr(VMALLOCUSED)
        page_tables = self._get_attr(PAGETABLES)
        kernel_stack = self._get_attr(KERNELSTACK)
        hw_corrupted = self._get_attr(HARDWARECORRUPTED)
        bounce = self._get_attr(BOUNCE)
        logger.debug("Kernel Used Memory Details, %s, %s, %s, %s, %s, %s",
                     slab, vmalloc_used, page_tables, kernel_stack,
                     hw_corrupted, bounce)
        used = [
            slab.value.value(),
            vmalloc_used.value.value(),
            page_tables.value.value(),
            kernel_stack.value.value(),
            hw_corrupted.value.value(),
            bounce.value.value(),
        ]
        kernel_used = 0
        for val in used:
            if isinstance(val, int):
                kernel_used += val
        kval = IntUnitValue(kernel_used, "kB")
        kval.fmt = DEF_FMT_2
        logger.debug("Total Kernel Used Memory: %d %s", kval.value(),
                     kval.unit())
        return Attr("KERNEL", kval)

    def _get_user_used_mem(self) -> Attr:
        """
        From http://linuxperf.com/?cat=7
        User Used Memory includes:
            1. Cached
            2. AnonPages
            3. Buffers
            4. HugePages_Total * Hugepagesize

        > 用户进程的内存主要有三种统计口径：
        >   1. 围绕LRU进行统计
        >     【(Active + Inactive + Unevictable) + (HugePages_Total * Hugepagesize)】
        >   2. 围绕Page Cache进行统计
        >     当SwapCached为0的时候，用户进程的内存总计如下：
        >     【(Cached + AnonPages + Buffers) + (HugePages_Total * Hugepagesize)】
        >     当SwapCached不为0的时候，以上公式不成立，因为SwapCached可能会含有Shmem，
        >     而Shmem本来被含在Cached中，一旦swap-out就从Cached转移到了SwapCached，
        >     可是我们又不能把SwapCached加进上述公式中，
        >     因为SwapCached虽然不与Cached重叠却与AnonPages有重叠，
        >     它既可能含有Shared memory又可能含有Anonymous Pages。
        >   3. 围绕RSS/PSS进行统计
        >     把/proc/[1-9]*/smaps 中的 Pss 累加起来就是所有用户进程占用的内存，
        >     但是还没有包括Page Cache中unmapped部分、以及HugePages，所以公式如下：
        >     ΣPss + (Cached – mapped) + Buffers + (HugePages_Total * Hugepagesize)
        """
        cached = self._get_attr(CACHED)
        anon_pages = self._get_attr(ANONPAGES)
        buffers = self._get_attr(BUFFERS)
        hp_total = self.get_attr_int_value(HUGEPAGES_TOTAL)
        hp_size = self.get_attr_int_value(HUGEPAGESIZE)
        logger.debug("User Used Memory Details, %s, %s, %s, %d, %d", cached,
                     anon_pages, buffers, hp_total, hp_size)
        used = [
            cached.value.value(),
            anon_pages.value.value(),
            buffers.value.value(),
            hp_total * hp_size,
        ]
        user_used = 0
        for val in used:
            if isinstance(val, int):
                user_used += val
        uval = IntUnitValue(user_used, "kB")
        uval.fmt = DEF_FMT_2
        logger.debug("Total User Used Memory: %d %s", uval.value(),
                     uval.unit())
        return Attr("USER", uval)

    def get_attrs(self, names: List[str]) -> List[Attr]:
        attrs = []
        attrs.append(current_time_attr())
        if not names:
            names = ["KERNEL", "USER", MEMFREE, MEMTOTAL]
        if names:
            for name in names:
                attr = self._get_attr(name)
                if attr and attr != EmptyAttr:
                    attrs.append(attr)
        return attrs

    def get_attr(self, name: str) -> Attr:
        return self._get_attr(name)


class Vmalloc(NamedTuple):
    vaddr_start: int
    vaddr_end: int
    size_in_bytes: int
    caller: str
    # optional fields
    pages: int = 0    # number of pages
    phys: int = 0    # physical address
    ioremap: bool = False    # I/O mapping (ioremap() and friends)
    vmalloc: bool = False    # vmalloc() area
    vmap: bool = False    # vmap()ed pages
    user: bool = False    # VM_USERMAP area
    vpages: bool = False    # buffer for pages pointers was vmalloced (huge area)
    npages_on_memory_node: Dict[str, int] = {}
    tag: str = ""


EmptyVmalloc = Vmalloc(0, 0, 0, "")


class VmallocInfo:

    PAGES_PAT = re.compile(r"^pages=(?P<pages>\d+)$")

    NODE_PAT = re.compile(r"^N(?P<index>\d)=(?P<pages>\d+)$")

    PHYS_PAT = re.compile(r"phys=(?P<addr>\d+)")

    TAG_PAT = re.compile(r"^(?P<tag>\[\w+\])$")

    UNPURGED = "unpurged"
    IOREMAP = "ioremap"
    VMALLOC = "vmalloc"
    VMAP = "vmap"
    USER = "user"
    VPAGES = "vpages"

    @staticmethod
    def _parse(line: str) -> Vmalloc:
        # pylint: disable=too-many-locals,too-many-statements
        # https://github.com/torvalds/linux/blob/master/Documentation/filesystems/proc.rst#vmallocinfo
        units = re.split(r"\s+", line)
        addr_range = units[0]
        idx = addr_range.find("-")
        start = int(addr_range[0:idx], base=16)
        end = int(addr_range[idx + 1:], base=16)
        size = int(units[1])
        if end - start != size:
            logger.error("end(%d) - start(%d) != size(%d)", end, start, size)
        caller = units[2]
        if caller == VmallocInfo.UNPURGED:
            return Vmalloc(start, end, size, VmallocInfo.UNPURGED)
        # parse optional
        pages = 0
        phys = 0
        ioremap = False
        vmalloc = False
        vmap = False
        user = False
        vpages = False
        npages_on_memory_node: Dict[str, int] = {}
        tag = ""
        for unit in units[3:]:
            match = VmallocInfo.TAG_PAT.match(unit)
            if match:
                tag = match.group("tag")
                continue
            match = VmallocInfo.PAGES_PAT.match(unit)
            if match:
                pages = int(match.group("pages"))
                continue
            match = VmallocInfo.PHYS_PAT.match(unit)
            if match:
                phys = int(match.group("addr"), base=16)
                continue
            if unit == VmallocInfo.IOREMAP:
                ioremap = True
                continue
            if unit == VmallocInfo.VMALLOC:
                vmalloc = True
                continue
            if unit == VmallocInfo.VMAP:
                vmap = True
                continue
            if unit == VmallocInfo.USER:
                user = True
                continue
            if unit == VmallocInfo.VPAGES:
                vpages = True
                continue
            match = VmallocInfo.NODE_PAT.match(unit)
            if match:
                node_idx = match.group("index")
                node_pages = match.group("pages")
                npages_on_memory_node[f"N{node_idx}"] = int(node_pages)
                continue
        return Vmalloc(start, end, size, caller, pages, phys, ioremap, vmalloc,
                       vmap, user, vpages, npages_on_memory_node, tag)

    def __init__(self, path="/proc/vmallocinfo"):
        with open_file(path) as vminfo:
            lines = [l.strip() for l in vminfo.readlines()]
        vms = []
        for line in lines:
            vmalloc = VmallocInfo._parse(line)
            if vmalloc != EmptyVmalloc:
                vms.append(vmalloc)
        self._vmallocs = vms
        callers = defaultdict(list)
        for vmalloc in vms:
            callers[vmalloc.caller].append(vmalloc)
        self._callers: Dict[str, List[Vmalloc]] = callers
        by_size: List[Tuple[str, int]] = []
        for cname, clist in self._callers.items():
            csize = 0
            for cvm in clist:
                csize += cvm.size_in_bytes
            by_size.append((cname, csize))
        self._sorted_by_size = sorted(by_size,
                                      key=lambda pair: pair[1],
                                      reverse=True)



from typing import NamedTuple, List

from xproc.value import (
    Attr,
    parse_list_int_val,
    ListIntValue,
)


class CPUStat(NamedTuple):
    user: int
    nice: int
    system: int
    idle: int
    iowait: int
    irq: int
    softirq: int
    steal: int
    guest: int
    guest_nice: int


EmptyCpuStat = CPUStat(0, 0, 0, 0, 0, 0, 0, 0, 0, 0)


class SoftIRQStat(NamedTuple):
    total: int
    hi: int
    timer: int
    net_tx: int
    net_rx: int
    block: int
    irq_poll: int
    tasklet: int
    sched: int
    hrtimer: int
    rcu: int


EmptySoftIRQStat = SoftIRQStat(0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0)


class SystemStat(NamedTuple):
    cpu: CPUStat
    cpus: List[CPUStat]
    intr: List[int]
    ctxt: int
    btime_in_sec: int
    processes: int
    procs_running: int
    procs_blocked: int
    softirq: SoftIRQStat


def _parse_cpustat(attr: Attr) -> CPUStat:
    if isinstance(attr.value, ListIntValue):
        return CPUStat(*attr.value.value())
    return EmptyCpuStat


def current_system_stat(path="/proc/stat"):
    with open(path, "r", encoding="utf-8") as stat:
        lines = stat.readlines()
    attrs: List[Attr] = []
    for line in lines:
        line = line.strip()
        idx = line.find(" ")
        name = line[0:idx]
        value = line[idx + 1:]
        attrs.append(parse_list_int_val(name, value))
    cpus = []
    d_other = {}
    i_other = {}
    softirq = []
    for attr in attrs:
        name = attr.name
        if name.startswith("cpu"):
            cpus.append(_parse_cpustat(attr))
        elif name.startswith("intr"):
            value = attr.value
            if isinstance(value, ListIntValue):
                d_other[name] = value.value()
        elif name.startswith("softirq"):
            value = attr.value
            if isinstance(value, ListIntValue):
                softirq = value.value()
        else:
            value = attr.value
            if isinstance(value, ListIntValue):
                i_other[name] = value.value()[0]
    return SystemStat(cpus[0], cpus[1:], d_other.get("intr"),
                      i_other.get("ctxt"), i_other.get("btime"),
                      i_other.get("processes"), i_other.get("procs_running"),
                      i_other.get("procs_blocked"), SoftIRQStat(*softirq))
